	.text
	.code 32
	.align  0

	.global start

	/* Stack Sizes */
	.set  UND_STACK_SIZE, 0x00000004
	.set  ABT_STACK_SIZE, 0x00000004
	.set  FIQ_STACK_SIZE, 0x00000200
	.set  IRQ_STACK_SIZE, 0X00000400
	.set  SVC_STACK_SIZE, 0x00000400

	/* Standard definitions of Mode bits and Interrupt (I & F) flags in PSRs */
	.set  MODE_USR, 0x10            /* User Mode */
	.set  MODE_FIQ, 0x11            /* FIQ Mode */
	.set  MODE_IRQ, 0x12            /* IRQ Mode */
	.set  MODE_SVC, 0x13            /* Supervisor Mode */
	.set  MODE_ABT, 0x17            /* Abort Mode */
	.set  MODE_UND, 0x1B            /* Undefined Mode */
	.set  MODE_SYS, 0x1F            /* System Mode */

	.set  I_BIT, 0x80               /* when I bit is set, IRQ is disabled */
	.set  F_BIT, 0x40               /* when F bit is set, FIQ is disabled */


start:
__start:
_mainCRTStartup:

	/* Clear BSS. */

	ldr     r1, =__bss_beg__        @ Start of memory block
	ldr     r3, =__bss_end__        @ End of memory block
	subs    r3, r3, r1              @ Calculate length of block
	beq     .end_clear_loop         @ If 0, nothing to do
	mov     r2, #0                  @ Fill value

.clear_loop:
	strb    r2, [r1], #1            @ Store byte, r1++
	subs    r3, r3, #1              @ Decrement counter
	bgt     .clear_loop             @ >= 0, go again

.end_clear_loop:


	/* Initialise data. */

	ldr     r1, =__data_beg__       @ Destination (.data in RAM)
	ldr     r2, =__data_beg_src__   @ Source (.data in FLASH)
	ldr     r3, =__data_end__       @ End of .data in RAM
	subs    r3, r3, r1              @ Calculate length of block
	beq     .end_set_loop           @ If 0, nothing to do

.set_loop:
	ldrb    r4, [r2], #1            @ Get byte from source, r2++
	strb    r4, [r1], #1            @ Store byte to destination, r1++
	subs    r3, r3, #1              @ Decrement loop counter
	bgt     .set_loop               @ >= 0, go again

.end_set_loop:


	/* Copy interrupt vectors */

	ldr     r1, =__intvects_beg__     @ Destination (.intvects in RAM)
	ldr     r2, =__intvects_beg_src__ @ Source (.intvects in FLASH)
	ldr     r3, =__intvects_end__     @ End of .intvects in RAM
	subs    r3, r3, r1                @ Calculate length of block
	beq     .end_ints_loop            @ If 0, nothing to do

.ints_loop:
	ldrb    r4, [r2], #1            @ Get byte from source, r2++
	strb    r4, [r1], #1            @ Store byte to destination, r1++
	subs    r3, r3, #1              @ Decrement loop counter
	bgt     .ints_loop              @ >= 0, go again

.end_ints_loop:


	/* Setup a stack for each mode - note that this only sets up a usable stack
	for system/user, SWI and IRQ modes.   Also each mode is setup with
	interrupts initially disabled. */

	ldr   r0, =__stack_end__            @ Get top of stack space
	msr   CPSR_c, #MODE_UND|I_BIT|F_BIT @ Undefined Instruction Mode
	mov   sp, r0                        @ Set undef mode SP

	sub   r0, r0, #UND_STACK_SIZE       @ Subtract undef stack size for abort stack start
	msr   CPSR_c, #MODE_ABT|I_BIT|F_BIT @ Abort Mode
	mov   sp, r0                        @ Set abort mode SP

	sub   r0, r0, #ABT_STACK_SIZE       @ Subtract abort stack size for FIQ stack start
	msr   CPSR_c, #MODE_FIQ|I_BIT|F_BIT @ FIQ Mode
	mov   sp, r0                        @ Set FIQ mode SP

	sub   r0, r0, #FIQ_STACK_SIZE       @ Subtract FIQ stack size for IRQ stack start
	msr   CPSR_c, #MODE_IRQ|I_BIT|F_BIT @ IRQ Mode
	mov   sp, r0                        @ Set IRQ mode SP

	sub   r0, r0, #IRQ_STACK_SIZE       @ Subtract IRQ stack size for SVC stack start
	msr   CPSR_c, #MODE_SVC|I_BIT|F_BIT @ Supervisor Mode
	mov   sp, r0                        @ Set supervisor mode SP

	sub   r0, r0, #SVC_STACK_SIZE       @ Subtract supervisor stack size for system/user stack start
	msr   CPSR_c, #MODE_SYS|I_BIT|F_BIT @ System Mode
	mov   sp, r0                        @ Set system/user mode SP

	/* We want to start in supervisor mode.  Operation will switch to system
	mode when the first task starts. */
	msr   CPSR_c, #MODE_SVC|I_BIT|F_BIT


	mov		r0, #0          /* no arguments  */
	mov		r1, #0          /* no argv either */

	bl		main


	/* Setup vector table. */

	.section .startup,"ax"
	.code 32
	.align 0

	ldr     pc, _start              @ reset - _start
	ldr     pc, _undf               @ undefined - _undf
	ldr     pc, _swi                @ SWI - _swi
	ldr     pc, _pabt               @ program abort - _pabt
	ldr     pc, _dabt               @ data abort - _dabt
	ldr     pc, _start              @ reserved - reset
	ldr     pc, [pc, #-0xff0]       @ IRQ - read VIC_VectAddr (magic!  It's at 0xfffff030)
	ldr     pc, _fiq                @ FIQ - _fiq

_start: .word __start                   /* reset                    */
_undf:  .word __undf                    /* undefined				*/
_swi:   .word __swi                     /* SWI						*/
_pabt:  .word __pabt                    /* program abort			*/
_dabt:  .word __dabt                    /* data abort				*/
_fiq:   .word __fiq                     /* FIQ						*/

__undf: b     .                         /* undefined				*/
__swi:  b     .                         /* SWI          			*/
__pabt: b     .                         /* program abort			*/
__dabt: b     .                         /* data abort				*/
__fiq:  b     .                         /* FIQ						*/


	.global __intvects
	.section .intvects,"ax"
	.align  0

__intvects:
	ldr     pc, _ram_reset          @ reset
	ldr     pc, _ram_undf           @ undefined
	ldr     pc, _ram_swi            @ SWI
	ldr     pc, _ram_pabt           @ program abort
	ldr     pc, _ram_dabt           @ data abort
	ldr     pc, _ram_rsrvd          @ reserved
	ldr     pc, [pc, #-0xff0]       @ IRQ - read VIC_VectAddr (magic!  It's at 0xfffff030)
	ldr     pc, _ram_fiq            @ FIQ

_ram_reset:   .word   __start                 @ reset
_ram_undf:    .word   __undf                  @ undefined
_ram_swi:     .word   __swi                   @ SWI
_ram_pabt:    .word   __pabt                  @ program abort
_ram_dabt:    .word   __dabt                  @ data abort
_ram_rsrvd:   .word   __start                 @ reserved
_ram_irq:     .word   0                       @ (not used, just reserved space)
_ram_fiq:     .word   __fiq                   @ FIQ
